@@ -16,51 +16,121 @@ from cvsps import CvspsWorkingDir
 from source import GetUpstreamChangesetsFailure
 from config import ConfigurationError
 
-def compare_cvs_revs(rev1, rev2):
-    """Compare two CVS revision numerically, not alphabetically."""
-
-    if not rev1: rev1 = '0'
-    if not rev2: rev2 = '0'
+def normalize_cvs_rev(rev):
+    """Convert a revision string to a tuple of numbers, eliminating the
+    penultimate zero in a 'magic branch number' if there is one.
+    1.1.1.1 is converted to (1,1). """
+    if not rev: rev = '0'
 
     # handle locked files by taking only the first part of the
     # revision string to handle gracefully lines like "1.1 locked"
-    rev1 = rev1.split(' ')[0]
-    rev2 = rev2.split(' ')[0]
-    r1 = [int(n) for n in rev1.split('.')]
-    r2 = [int(n) for n in rev2.split('.')]
+    rev = rev.split(' ')[0]
+
+    r = [int(n) for n in rev.split('.')]
+    # convert "magic branch numbers" like 1.2.0.2 to regular
+    # branch numbers like 1.2.2.
+    if len(r) > 2 and r[-2] == 0:
+        r = r[0:-2] + r[-1:]
+
+    if r == [1,1,1,1]:
+        r = [1,1]
+
+    return tuple(r)
+
+def compare_cvs_revs(revstr1, revstr2):
+    """Compare two CVS revision strings numerically, not alphabetically."""
+
+    r1 = normalize_cvs_rev(revstr1)
+    r2 = normalize_cvs_rev(revstr2)
 
     return cmp(r1, r2)
 
+def cvs_revs_same_branch(rev1, rev2):
+    """True iff the two normalized revision numbers are on the same branch."""
 
-def changesets_from_cvslog(log, module):
+    # Odd-length revisions are branch numbers, even-length ones
+    # are revision numbers.
+
+    # Two branch numbers can't be on the same branch unless they're identical.
+    if len(rev1) % 2 and len(rev2) % 2:
+        return rev1 == rev2
+
+    # Two revision numbers are on the same branch if they
+    # agree up to the last number.
+    if len(rev1) % 2 == 0 and len(rev2) % 2 == 0:
+        return rev1[0:-1] == rev2[0:-1]
+
+    # One branch number, one revision number.  If by removing the last number
+    # of one you get the other, then they're on the same branch, regardless of
+    # which is longer.  E.g. revision 1.2 is the root of the branch 1.2.2;
+    # revision 1.2.2.2 is directly on the branch 1.2.2.
+    if rev1[0:-1] == rev2:
+        return True
+
+    if rev2[0:-1] == rev1:
+        return True
+
+    return False
+
+def is_branch(rev):
+    """True iff the given (normalized) revision number is a branch number"""
+    if len(rev) % 2:
+        return True
+
+def rev2branch(rev):
+    """Return the branch on which this (normalized) revision lies"""
+    assert not is_branch(rev)
+    return rev[0:-1]
+
+
+def changesets_from_cvslog(log, module, branch=None, entries=None, since=None):
     """
     Parse CVS log.
     """
 
     from datetime import timedelta
 
-    collected = ChangeSetCollector(log, module)
-    collapsed = []
+    collected = ChangeSetCollector(log, module, branch, entries, since)
 
     threshold = timedelta(seconds=180)
     last = None
 
+    # Loop over collected changesets, and collapse those with same author,
+    # same changelog and that were committed within 3 minutes one from the
+    # other. If they have entries in common, keep them separated. Special
+    # treatment to deleted entries, given that sometime there are two
+    # deletions on the same file: in that case, keep only the last one,
+    # with higher revision.
     for cs in collected:
-        if (last and last.author == cs.author and  last.log == cs.log and
+        if (last and last.author == cs.author and last.log == cs.log and
             abs(lastts - cs.date) < threshold and
+            not last.tags and
             not [e for e in cs.entries
-                 if e.name in [n.name for n in last.entries]]):
-            last.entries.extend(cs.entries)
+                 if e.name in [n.name for n in last.entries
+                               if n.action_kind <> e.action_kind]]):
+            for e in cs.entries:
+                if e.action_kind == e.DELETED:
+                    doubledelete = False
+                    for n in last.entries:
+                        if n.name == e.name and n.action_kind == n.DELETED:
+                            doubledelete = True
+                            n.new_revision = e.new_revision
+                            break
+                    if not doubledelete:
+                        last.entries.append(e)
+                else:
+                    last.entries.append(e)
             if lastts < cs.date:
                 lastts = cs.date
         else:
             if last:
                 last.date = lastts
+                yield last
             last = cs
             lastts = cs.date
-            collapsed.append(cs)
 
-    return collapsed
+    if last:
+        yield last
 
 
 def _getGlobalCVSRevision(timestamp, author):
@@ -92,23 +162,30 @@ class ChangeSetCollector(object):
     intra_sep = '-' * 28 + '\n'
     inter_sep = '=' * 77 + '\n'
 
-    def __init__(self, log, module):
+    def __init__(self, log, module, branch, entries, since):
         """
         Initialize a ChangeSetCollector instance.
 
         Loop over the modified entries and collect their logs.
         """
 
+        from logging import getLogger
+
         self.changesets = {}
         """The dictionary mapping (date, author, log) to each entry."""
 
-        self.log = log
+        self.cvslog = log
         """The log to be parsed."""
 
         self.module = module
         """The CVS module name."""
 
-        self.__parseCvsLog()
+        self.__lookahead = []
+        """The look ahead line stack."""
+
+        self.log = getLogger('tailor.vcpx.cvs.collector')
+
+        self.__parseCvsLog(branch, entries, since)
 
     def __iter__(self):
         keys = self.changesets.keys()
@@ -122,14 +199,18 @@ class ChangeSetCollector(object):
 
         key = (timestamp, author, changelog)
         if self.changesets.has_key(key):
-            return self.changesets[key].addEntry(entry, revision)
+            cs = self.changesets[key]
+            for e in cs.entries:
+                if e.name == entry:
+                    return e
+            return cs.addEntry(entry, revision)
         else:
             cs = Changeset(_getGlobalCVSRevision(timestamp, author),
                            timestamp, author, changelog)
             self.changesets[key] = cs
             return cs.addEntry(entry, revision)
 
-    def __readline(self):
+    def __readline(self, lookahead=False):
         """
         Read a line from the log, intercepting the directory being listed.
 
@@ -137,7 +218,15 @@ class ChangeSetCollector(object):
         the root of the working copy.
         """
 
-        l = self.log.readline()
+        if lookahead:
+            l = self.cvslog.readline()
+            self.__lookahead.append(l)
+            return l
+        else:
+            if self.__lookahead:
+                l = self.__lookahead.pop(0)
+            else:
+                l = self.cvslog.readline()
         while l.startswith('cvs rlog: Logging '):
             currentdir = l[18:-1]
             # If the directory starts with the module name, keep
@@ -151,7 +240,7 @@ class ChangeSetCollector(object):
                     self.__currentdir = currentdir[slash+1:]
                 else:
                     self.__currentdir = ''
-            l = self.log.readline()
+            l = self.cvslog.readline()
 
         return l
 
@@ -214,8 +303,14 @@ class ChangeSetCollector(object):
             l = self.__readline()
 
         mesg = []
-        while l not in (None, '', self.inter_sep, self.intra_sep):
-            mesg.append(l[:-1])
+        while True:
+            if l == self.intra_sep:
+                if self.__readline(True).startswith('revision '):
+                    break
+            if l in (None, '', self.inter_sep):
+                break
+            if l<>self.intra_sep:
+                mesg.append(l[:-1])
             l = self.__readline()
 
         if len(mesg)==1 and mesg[0] == '*** empty log message ***':
@@ -225,16 +320,23 @@ class ChangeSetCollector(object):
 
         return (date, author, changelog, entry, rev, state, newentry)
 
-    def __parseCvsLog(self):
+    def __parseCvsLog(self, branch, entries, since):
         """Parse a complete CVS log."""
 
+        from changes import Changeset
         from os.path import split, join
         import sre
+        from datetime import timedelta
+        from time import strptime
+        from datetime import datetime
 
         revcount_regex = sre.compile('\\bselected revisions:\\s*(\\d+)\\b')
 
         self.__currentdir = None
 
+        file2rev2tags = {}
+        tagcounts = {}
+        branchnum = None
         while 1:
             l = self.__readline()
             while l and not l.startswith('RCS file: '):
@@ -247,20 +349,54 @@ class ChangeSetCollector(object):
                    "Missed 'cvs rlog: Logging XX' line"
 
             entry = join(self.__currentdir, split(l[10:-1])[1][:-2])
+            if entries is not None:
+                while l and not l.startswith('head: '):
+                    l = self.__readline()
+                assert l, "Missed 'head:' line"
+                if branch is None:
+                    branchnum = normalize_cvs_rev(l[6:-1])
+                    branchnum = rev2branch(branchnum)
+
+                while l and not l == 'symbolic names:\n':
+                    l = self.__readline()
+
+                assert l, "Missed 'symbolic names:' line"
+
+                l = self.__readline()
+                rev2tags = {}
+                while l.startswith('\t'):
+                    tag,revision = l[1:-1].split(': ')
+                    tagcounts[tag] = tagcounts.get(tag,0) + 1
+                    revision = normalize_cvs_rev(revision)
+                    rev2tags.setdefault(revision,[]).append(tag)
+                    if tag == branch:
+                        branchnum = revision
+
+                    l = self.__readline()
+
+                # branchnum may still be None, if this file doesn't exist
+                # on the requested branch.
+
+                # filter out branch tags, and tags for revisions that are
+                # on other branches.
+                for revision in rev2tags.keys():
+                    if is_branch(revision) or \
+                       not branchnum or \
+                       not cvs_revs_same_branch(revision,branchnum):
+                        del rev2tags[revision]
+
+                file2rev2tags[entry] = rev2tags
 
             expected_revisions = None
-            while 1:
-                l = self.__readline()
-                if l in (self.inter_sep, self.intra_sep):
-                    break
-
+            while l not in (self.inter_sep, self.intra_sep):
                 m = revcount_regex.search(l)
                 if m is not None:
                     expected_revisions = int(m.group(1))
-
+                l = self.__readline()
             last = previous = None
             found_revisions = 0
-            while l <> self.inter_sep:
+            while (l <> self.inter_sep or
+                   not self.__readline(True).startswith('revision ')):
                 cs = self.__parseRevision(entry)
                 if cs is None:
                     break
@@ -280,15 +416,108 @@ class ChangeSetCollector(object):
                 found_revisions = found_revisions + 1
 
                 if previous and last.action_kind == last.DELETED:
-                    previous.action_kind = previous.ADDED
+                    # For unknown reasons, sometimes there are two dead
+                    # revision is a row.
+                    if previous.action_kind <> last.DELETED:
+                        previous.action_kind = previous.ADDED
 
                 previous = last
 
             if expected_revisions <> found_revisions:
-                print 'warning: expecting %s revisions, read %s revisions' % \
-                      ( expected_revisions, found_revisions )
+                self.log.warning('Expecting %s revisions, found %s',
+                                 expected_revisions, found_revisions)
 
-    # end of __parseCvsLog()
+        # If entries is not given, don't try to desume tags information
+        if entries is None:
+            return
+
+        # Determine the current revision of each live
+        # (i.e. non-deleted) entry.
+        state = dict(entries.getFileVersions())
+
+        # before stepping through changes, see if the initial state is
+        # taggable.  If so, add an initial changeset that does nothing
+        # but tag, using the date of the last revision tailor imported
+        # on its previous run.  There's no way to tell when the tag
+        # was really applied, so we don't know if it was seen on the
+        # last run or not.  Before applying the tag on the other end,
+        # we'll have to check whether it's already been applied.
+        tags = self.__getApplicableTags(state, file2rev2tags, tagcounts)
+        if tags:
+            if since == None:
+                # I think this could only happen if the CVS repo was
+                # tagged before any files were added to it.  We could
+                # probably get a better date by looking at when the
+                # files were added, but who cares.
+                timestamp = datetime(1900,1,1)
+            else:
+                # "since" is a revision name read from the state file,
+                # which means it was originally generated by
+                # getGlobalCVSRevision.  The format string "%Y-%m-%d
+                # %H:%M:%S" matches the format generated by the implicit
+                # call to timestamp.__str__() in getGlobalCVSRevision.
+                y,m,d,hh,mm,ss,d1,d2,d3 = strptime(since, "%Y-%m-%d %H:%M:%S")
+                timestamp = datetime(y,m,d,hh,mm,ss)
+            author = "unknown tagger"
+            changelog = "tag %s %s" % (timestamp, tags)
+            key = (timestamp, author, changelog)
+            self.changesets[key] = Changeset(_getGlobalCVSRevision(timestamp,
+                                                                   author),
+                                             timestamp,author,changelog,
+                                             tags=tags)
+
+        # Walk through the changesets, identifying ones that result in
+        # a state with a tag.  Add that info to the changeset.
+        for cs in self.__iter__():
+            self.__updateState(state, cs)
+            cs.tags = self.__getApplicableTags(state, file2rev2tags, tagcounts)
+
+    def __getApplicableTags(self,state,taginfo,expectedcounts):
+        # state:   a dictionary mapping filename->revision
+        #
+        # taginfo: a two-level dictionary mapping
+        #          tagname->revision->list of tags.
+        #
+        # expectedcounts: a dictionary mapping tagname->number of
+        #                 files tagged with that name.
+        observedcounts = {}
+        possibletags = []
+        for filename, revno in state.iteritems():
+            filetags = taginfo[filename].get(revno,[])
+            if len(possibletags) == 0:
+                # first iteration of loop
+                possibletags = filetags
+
+            # Intersection of possibletags and filetags.  I'm
+            # avoiding using python sets to preserve python 2.3
+            # compatibility.
+            possibletags = [t for t in possibletags if t in filetags]
+            for t in filetags:
+                 observedcounts[t] = observedcounts.get(t,0) + 1
+
+            if len(possibletags) == 0:
+                break
+
+        # All currently existing files carry the tags in possibletags.
+        # But that doesn't mean that the tags correspond to this
+        # state--we might need to create additional files before
+        # tagging.
+        possibletags = [t for t in possibletags if
+                        observedcounts[t] == expectedcounts[t]]
+
+        return possibletags
+
+    def __updateState(self,state, changeset):
+        for e in changeset.entries:
+            if e.action_kind in (e.ADDED, e.UPDATED):
+                state[e.name] = normalize_cvs_rev(e.new_revision)
+            elif e.action_kind == e.DELETED:
+                if state.has_key(e.name):
+                    del state[e.name]
+            elif e.action_kind == e.RENAMED:
+                if state.has_key(e.name):
+                    del state[e.old_name]
+                state[e.name] = normalize_cvs_rev(e.new_revision)
 
 
 class CvsWorkingDir(CvspsWorkingDir):
@@ -300,13 +529,14 @@ class CvsWorkingDir(CvspsWorkingDir):
     def _getUpstreamChangesets(self, sincerev):
         from os.path import join, exists
         from datetime import timedelta
+        from time import sleep
 
         from codecs import getreader
 
         try:
             reader = getreader(self.repository.encoding)
-        except LookupError, err:
-            raise ConfigurationError('Encoding %r does not seem to be '
+        except (ValueError, LookupError), err:
+            raise ConfigurationError('Encoding "%s" does not seem to be '
                                      'allowed on this system (%s): you '
                                      'may override the default with '
                                      'something like "encoding = ascii" in '
@@ -314,15 +544,14 @@ class CvsWorkingDir(CvspsWorkingDir):
                                      (self.repository.encoding, err,
                                       self.repository.name))
 
-        branch = ''
+        branch = None
         fname = join(self.basedir, 'CVS', 'Tag')
         if exists(fname):
             tag = open(fname).read()
             if tag[0] == 'T':
                 branch=tag[1:-1]
 
-        cmd = self.repository.command("-f", "-d", "%(repository)s", "rlog",
-                                      "-N")
+        cmd = self.repository.command("-f", "-d", "%(repository)s", "rlog")
 
         if not sincerev or sincerev in ("INITIAL", "HEAD"):
             # We are bootstrapping, trying to collimate the actual
@@ -358,16 +587,33 @@ class CvsWorkingDir(CvspsWorkingDir):
 
         cvslog = ExternalCommand(command=cmd)
 
-        log = cvslog.execute(self.repository.module, stdout=PIPE, stderr=STDOUT,
-                             repository=self.repository.repository,
-                             since=since, branch=branch or 'HEAD', TZ='UTC')[0]
+        retry = 0
+        while True:
+            log = cvslog.execute(self.repository.module, stdout=PIPE,
+                                 stderr=STDOUT, since=since,
+                                 repository=self.repository.repository,
+                                 branch=branch or 'HEAD', TZ='UTC')[0]
+            if cvslog.exit_status:
+                retry += 1
+                if retry>3:
+                    break
+                delay = 2**retry
+                self.log.info("%s returned status %s, "
+                              "retrying in %d seconds...",
+                              str(cvslog), cvslog.exit_status,
+                              delay)
+                sleep(retry)
+            else:
+                break
 
         if cvslog.exit_status:
             raise GetUpstreamChangesetsFailure(
                 "%s returned status %d" % (str(cvslog), cvslog.exit_status))
 
         log = reader(log)
-        return changesets_from_cvslog(log, self.repository.module)
+        return changesets_from_cvslog(log, self.repository.module, branch,
+                                      CvsEntries(self.repository.rootdir),
+                                      since)
 
     def _checkoutUpstreamRevision(self, revision):
         """
@@ -504,3 +750,18 @@ class CvsEntries(object):
                 latest = e
 
         return latest
+
+    def getFileVersions(self, prefix=''):
+        """Return a set of (entry name, version number) pairs."""
+
+        from os.path import join
+
+        pairs = [(prefix+e.filename, normalize_cvs_rev(e.cvs_version))
+                 for e in self.files.values()]
+
+        for dirname, entries in self.directories.iteritems():
+            pairs += [(prefix+filename, version)
+                      for filename, version in
+                      entries.getFileVersions("%s/" % dirname)]
+
+        return pairs
