@@ -72,16 +72,17 @@ def changesets_from_cvsps(log, sincerev=None):
             msg.append(l)
             l = log.readline()
 
-        pset['log'] = ''.join(msg)
-
         assert l.startswith('Members:'), "Parse error: %s" % l
 
-        pset['entries'] = entries = []
+        entries = []
         l = log.readline()
         seen = {}
         while l.startswith('\t'):
             if not sincerev or (sincerev<int(pset['revision'])):
-                file,revs = l[1:-1].split(':')
+                # Cannot use split here, file may contain ':'
+                cpos = l.rindex(':')
+                file = l[1:cpos]
+                revs = l[cpos+1:-1]
                 fromrev,torev = revs.strip().split('->')
 
                 # Due to the fuzzy mechanism, cvsps may group
@@ -126,7 +127,8 @@ def changesets_from_cvsps(log, sincerev=None):
             timestamp = datetime(y, m, d, hh, mm, ss)
             pset['date'] = timestamp
 
-            yield Changeset(**pset)
+            yield Changeset(pset['revision'], timestamp, pset['author'],
+                            ''.join(msg), entries)
 
 
 class CvspsWorkingDir(UpdatableSourceWorkingDir,
@@ -157,7 +159,6 @@ class CvspsWorkingDir(UpdatableSourceWorkingDir,
         if sincerev:
             sincerev = int(sincerev)
 
-        changesets = []
         cmd = self.repository.command("--cvs-direct", "-u", "-b", branch,
                                       "--root", self.repository.repository,
                                       cvsps=True)
@@ -165,9 +166,7 @@ class CvspsWorkingDir(UpdatableSourceWorkingDir,
         log = cvsps.execute(self.repository.module, stdout=PIPE, TZ='UTC')[0]
 
         for cs in changesets_from_cvsps(log, sincerev):
-            changesets.append(cs)
-
-        return changesets
+            yield cs
 
     def __maybeDeleteDirectory(self, entrydir, changeset):
         from os.path import join, exists
@@ -194,18 +193,18 @@ class CvspsWorkingDir(UpdatableSourceWorkingDir,
             if e.action_kind == e.UPDATED:
                 info = entries.getFileInfo(e.name)
                 if not info:
-                    self.log_info("promoting '%s' to ADDED at "
-                                  "revision %s" % (e.name, e.new_revision))
+                    self.log.debug('promoting "%s" to ADDED at '
+                                   'revision %s', e.name, e.new_revision)
                     e.action_kind = e.ADDED
                     self.__createParentCVSDirectories(changeset, e.name)
                 elif info.cvs_version == e.new_revision:
-                    self.log_info("skipping '%s' since it's already "
-                                  "at revision %s" % (e.name, e.new_revision))
+                    self.log.debug('skipping "%s" since it is already '
+                                   'at revision %s', e.name, e.new_revision)
                     continue
             elif e.action_kind == e.DELETED:
                 if not exists(join(self.basedir, e.name)):
-                    self.log_info("skipping '%s' since it's already "
-                                  "deleted" % e.name)
+                    self.log.debug('skipping "%s" since it is already '
+                                   'deleted', e.name)
                     self.__maybeDeleteDirectory(split(e.name)[0], changeset)
                     continue
             elif e.action_kind == e.ADDED and e.new_revision is None:
@@ -235,10 +234,9 @@ class CvspsWorkingDir(UpdatableSourceWorkingDir,
                         if retry>3:
                             break
                         delay = 2**retry
-                        self.log_info("%s returned status %s, "
-                                      "retrying in %d seconds..." %
-                                      (str(cvsup), cvsup.exit_status,
-                                       delay))
+                        self.log.warning("%s returned status %s, "
+                                         "retrying in %d seconds...",
+                                         str(cvsup), cvsup.exit_status, delay)
                         sleep(retry)
                     else:
                         break
@@ -248,7 +246,7 @@ class CvspsWorkingDir(UpdatableSourceWorkingDir,
                         "%s returned status %s" % (str(cvsup),
                                                    cvsup.exit_status))
 
-                self.log_info("%s updated to %s" % (e.name, e.new_revision))
+                self.log.debug("%s updated to %s", e.name, e.new_revision)
 
             if e.action_kind == e.DELETED:
                 self.__maybeDeleteDirectory(split(e.name)[0], changeset)
@@ -262,8 +260,9 @@ class CvspsWorkingDir(UpdatableSourceWorkingDir,
         Return the last applied changeset.
         """
 
-        from os.path import join, exists
+        from os.path import join, exists, split
         from cvs import CvsEntries, compare_cvs_revs
+        from time import sleep
 
         if not self.repository.module:
             raise InvocationError("Must specify a module name")
@@ -280,38 +279,53 @@ class CvspsWorkingDir(UpdatableSourceWorkingDir,
             elif ' ' in revision:
                 revision, timestamp = revision.split(' ', 1)
 
-        # Trasform the whole history in a list, since we need to
-        # visit it beginning from the last element
-        csets = list(self.getPendingChangesets(revision))
+        csets = self.getPendingChangesets(revision)
         if not csets:
             raise TargetInitializationFailure(
                 "Something went wrong: there are no changesets since "
                 "revision '%s'" % revision)
         if timestamp == 'INITIAL':
-            cset = csets[0]
-            timestamp = cset.date.isoformat(sep=' ')
+            initialcset = csets.next()
+            timestamp = initialcset.date.isoformat(sep=' ')
         else:
-            cset = None
+            initialcset = None
 
         if not exists(join(self.basedir, 'CVS')):
+            # CVS does not handle "checkout -d multi/level/subdir", so
+            # split the basedir and use it's parentdir as cwd below.
+            parentdir, subdir = split(self.basedir)
             cmd = self.repository.command("-q",
                                           "-d", self.repository.repository,
                                           "checkout",
-                                          "-d", self.repository.subdir)
+                                          "-d", subdir)
             if revision:
                 cmd.extend(["-r", revision])
             if timestamp:
                 cmd.extend(["-D", "%s UTC" % timestamp])
 
-            checkout = ExternalCommand(cwd=self.repository.rootdir, command=cmd)
-            checkout.execute(self.repository.module)
+            checkout = ExternalCommand(cwd=parentdir, command=cmd)
+            retry = 0
+            while True:
+                checkout.execute(self.repository.module)
+                if checkout.exit_status:
+                    retry += 1
+                    if retry>3:
+                        break
+                    delay = 2**retry
+                    self.log.warning("%s returned status %s, "
+                                     "retrying in %d seconds...",
+                                     str(checkout), checkout.exit_status,
+                                     delay)
+                    sleep(retry)
+                else:
+                    break
 
             if checkout.exit_status:
                 raise TargetInitializationFailure(
                     "%s returned status %s" % (str(checkout),
                                                checkout.exit_status))
         else:
-            self.log_info("Using existing %s" % self.basedir)
+            self.log.info("Using existing %s", self.basedir)
 
         if self.repository.tag_entries:
             self.__forceTagOnEachEntry()
@@ -328,32 +342,38 @@ class CvspsWorkingDir(UpdatableSourceWorkingDir,
         # out the actual cvsps revision
 
         found = False
-        if cset is None and csets:
-            cset = csets.pop()
-        while cset is not None:
-            for m in cset.entries:
+        csets = self.state_file.reversed()
+
+        def already_applied(cs, entries=entries):
+            "Loop over changeset entries to determine if it's already applied."
+
+            applied = False
+            for m in cs.entries:
                 info = entries.getFileInfo(m.name)
                 if info:
-                    actualversion = info.cvs_version
-                    found = compare_cvs_revs(actualversion,m.new_revision) >= 0
-                    if not found:
+                    odversion = info.cvs_version
+                    applied = compare_cvs_revs(odversion, m.new_revision) >= 0
+                    if not applied:
                         break
+            return applied
 
+        for cset in csets:
+            found = already_applied(cset)
             if found:
                 last = cset
                 break
 
-            if csets:
-                cset = csets.pop()
-            else:
-                cset = None
+        if not found and initialcset:
+            found = already_applied(initialcset)
+            if found:
+                last = initialcset
 
         if not found:
             raise TargetInitializationFailure(
                 "Something went wrong: unable to determine the exact upstream "
                 "revision of the checked out tree in '%s'" % self.basedir)
         else:
-            self.log_info("working copy up to cvs revision %s" % last.revision)
+            self.log.info("Working copy up to revision %s", last.revision)
 
         return last
 
@@ -454,11 +474,17 @@ class CvspsWorkingDir(UpdatableSourceWorkingDir,
         """
 
         from os import walk, rename
-        from os.path import join
+        from os.path import join, exists
 
         for dir, subdirs, files in walk(self.basedir):
             if dir[-3:] == 'CVS':
                 efn = join(dir, 'Entries')
+
+                # Strangeness is a foreign word in CVS: sometime
+                # the Entries isn't there...
+                if not exists(efn):
+                    continue
+
                 f = open(efn)
                 entries = f.readlines()
                 f.close()
@@ -497,9 +523,9 @@ class CvspsWorkingDir(UpdatableSourceWorkingDir,
         """
 
         from shwrap import ReopenableNamedTemporaryFile
-        from sys import getdefaultencoding
+        from locale import getpreferredencoding
 
-        encoding = ExternalCommand.FORCE_ENCODING or getdefaultencoding()
+        encoding = ExternalCommand.FORCE_ENCODING or getpreferredencoding()
 
         logmessage = []
         if patchname:
@@ -541,3 +567,15 @@ class CvspsWorkingDir(UpdatableSourceWorkingDir,
 
         self._removePathnames([oldname])
         self._addPathnames([newname])
+
+    def _tag(self, tagname):
+        """
+        Apply a tag.
+        """
+
+        cmd = self.repository.command("tag")
+        c = ExternalCommand(cwd=self.basedir, command=cmd)
+        c.execute(tagname)
+        if c.exit_status:
+            raise ChangesetApplicationFailure("%s returned status %d" %
+                                              (str(c), c.exit_status))
