@@ -22,15 +22,39 @@ Tailorized equivalent of
 %s
 """
 
-def changesets_from_darcschanges(changes, unidiff=False, repodir=None):
+def changesets_from_darcschanges(changes, unidiff=False, repodir=None,
+                                 chunksize=2**15):
     """
     Parse XML output of ``darcs changes``.
 
     Return a list of ``Changeset`` instances.
+
+    Filters out the (currently incorrect) tag info from
+    changesets_from_darcschanges_unsafe.
     """
 
-    from xml.sax import parse
-    from xml.sax.handler import ContentHandler
+    csets = changesets_from_darcschanges_unsafe(changes, unidiff,
+                                                repodir, chunksize)
+    for cs in csets:
+        cs.tags = None
+        yield cs
+
+def changesets_from_darcschanges_unsafe(changes, unidiff=False, repodir=None,
+                                        chunksize=2**15):
+    """
+    Do the real work of parsing the change log, including tags.
+    Warning: the tag information in the changsets returned by this
+    function are only correct if each darcs tag in the repo depends on
+    all of the patches that precede it.  This is not a valid
+    assumption in general--a tag that does not depend on patch P can
+    be pulled in from another darcs repo after P.  We collect the tag
+    info anyway because DarcsWorkingDir._currentTags() can use it
+    safely despite this problem.  Hopefully the problem will
+    eventually be fixed and this function can be renamed
+    changesets_from_darcschanges.
+    """
+    from xml.sax import make_parser
+    from xml.sax.handler import ContentHandler, ErrorHandler
     from changes import ChangesetEntry, Changeset
     from datetime import datetime
 
@@ -84,7 +108,8 @@ def changesets_from_darcschanges(changes, unidiff=False, repodir=None):
                                  self.current['date'],
                                  self.current['author'],
                                  changelog,
-                                 self.current['entries'])
+                                 self.current['entries'],
+                                 tags=self.current.get('tags',[]))
                 cset.darcs_hash = self.current['hash']
                 if self.darcsdiff:
                     cset.unidiff = self.darcsdiff.execute(
@@ -93,7 +118,10 @@ def changesets_from_darcschanges(changes, unidiff=False, repodir=None):
                 self.changesets.append(cset)
                 self.current = None
             elif name in ['name', 'comment']:
-                self.current[name] = ''.join(self.current_field)
+                val = ''.join(self.current_field)
+                if val[:4] == 'TAG ':
+                    self.current.setdefault('tags',[]).append(val[4:])
+                self.current[name] = val
             elif name == 'move':
                 entry = ChangesetEntry(self.new_name)
                 entry.action_kind = entry.RENAMED
@@ -114,16 +142,21 @@ def changesets_from_darcschanges(changes, unidiff=False, repodir=None):
         def characters(self, data):
             self.current_field.append(data)
 
-
+    parser = make_parser()
     handler = DarcsXMLChangesHandler()
-    parse(changes, handler)
-    changesets = handler.changesets
-
-    # sort changeset by date
-    changesets.sort(lambda x, y: cmp(x.date, y.date))
-
-    return changesets
+    parser.setContentHandler(handler)
+    parser.setErrorHandler(ErrorHandler())
 
+    chunk = changes.read(chunksize)
+    while chunk:
+        parser.feed(chunk)
+        for cs in handler.changesets:
+            yield cs
+        handler.changesets = []
+        chunk = changes.read(chunksize)
+    parser.close()
+    for cs in handler.changesets:
+        yield cs
 
 class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
     """
@@ -149,7 +182,7 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
 
         if pull.exit_status:
             raise GetUpstreamChangesetsFailure(
-                "%s returned status %d saying \"%s\"" %
+                "%s returned status %d saying\n%s" %
                 (str(pull), pull.exit_status, output.read()))
 
         l = output.readline()
@@ -157,8 +190,6 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
                          l == 'No remote changes to pull in!\n'):
             l = output.readline()
 
-        changesets = []
-
         if l <> 'No remote changes to pull in!\n':
             ## Sat Jul 17 01:22:08 CEST 2004  lele@nautilus
             ##   * Refix _getUpstreamChangesets for darcs
@@ -206,13 +237,16 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
                 cset.darcs_hash = '%s-%s-%s.gz' % (compactdate,
                                                    new(author).hexdigest()[:5],
                                                    phash.hexdigest())
-                changesets.append(cset)
+
+                if name.startswith('tagged'):
+                    self.log.warning("Skipping tag %s because I don't "
+                                     "propagate tags from darcs.", name)
+                else:
+                    yield cset
 
                 while not l.strip():
                     l = output.readline()
 
-        return changesets
-
     def _applyChangeset(self, changeset):
         """
         Do the actual work of applying the changeset to the working copy.
@@ -248,11 +282,11 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
             cmd.extend(['--patches', escape(changeset.revision)])
 
         pull = ExternalCommand(cwd=self.basedir, command=cmd)
-        output = pull.execute(stdout=PIPE, stderr=STDOUT)[0]
+        output = pull.execute(stdout=PIPE, stderr=STDOUT, input='y')[0]
 
         if pull.exit_status:
             raise ChangesetApplicationFailure(
-                "%s returned status %d saying \"%s\"" %
+                "%s returned status %d saying\n%s" %
                 (str(pull), pull.exit_status, output.read()))
 
         conflicts = []
@@ -260,8 +294,8 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
         while line:
             if line.startswith('We have conflicts in the following files:'):
                 files = output.readline()[:-1].split('./')[1:]
-                self.log_info("Conflict after 'darcs pull': '%s'" %
-                              ' '.join(files))
+                self.log.warning("Conflict after 'darcs pull': %s",
+                                 ' '.join(files))
                 conflicts.extend(['./' + f for f in files])
             line = output.readline()
 
@@ -270,7 +304,7 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
         changes = ExternalCommand(cwd=self.basedir, command=cmd)
         last = changesets_from_darcschanges(changes.execute(stdout=PIPE)[0])
         if last:
-            changeset.entries.extend(last[0].entries)
+            changeset.entries.extend(last.next().entries)
 
         return conflicts
 
@@ -283,7 +317,7 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
         be none of them in tailor context.
         """
 
-        self.log_info("Reverting changes to '%s', to solve the conflict" %
+        self.log.info("Reverting changes to %s, to solve the conflict",
                       ' '.join(conflict))
         cmd = self.repository.command("revert", "--all")
         revert = ExternalCommand(cwd=self.basedir, command=cmd)
@@ -302,18 +336,19 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
         if revision == 'INITIAL':
             initial = True
             cmd = self.repository.command("changes", "--xml-output",
-                                          "--repo", self.repository.repository)
+                                          "--repo", self.repository.repository,
+                                           "--reverse")
             changes = ExternalCommand(command=cmd)
             output = changes.execute(stdout=PIPE, stderr=STDOUT)[0]
 
             if changes.exit_status:
                 raise ChangesetApplicationFailure(
-                    "%s returned status %d saying \"%s\"" %
+                    "%s returned status %d saying\n%s" %
                     (str(changes), changes.exit_status,
                      output and output.read() or ''))
 
             csets = changesets_from_darcschanges(output)
-            changeset = csets[0]
+            changeset = csets.next()
 
             revision = 'hash %s' % changeset.darcs_hash
         else:
@@ -344,7 +379,7 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
 
                 if dpull.exit_status:
                     raise TargetInitializationFailure(
-                        "%s returned status %d saying \"%s\"" %
+                        "%s returned status %d saying\n%s" %
                         (str(dpull), dpull.exit_status, output.read()))
         else:
             # Use much faster 'darcs get'
@@ -359,7 +394,7 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
 
             if dget.exit_status:
                 raise TargetInitializationFailure(
-                    "%s returned status %d saying \"%s\"" %
+                    "%s returned status %d saying\n%s" %
                     (str(dget), dget.exit_status, output.read()))
 
         cmd = self.repository.command("changes", "--last", "1",
@@ -369,19 +404,19 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
 
         if changes.exit_status:
             raise ChangesetApplicationFailure(
-                "%s returned status %d saying \"%s\"" %
+                "%s returned status %d saying\n%s" %
                 (str(changes), changes.exit_status, output.read()))
 
         last = changesets_from_darcschanges(output)
 
-        return last[0]
+        return last.next()
 
 
     ## SyncronizableTargetWorkingDir
 
     def _addPathnames(self, names):
         """
-        Add some new filesystems objects.
+        Add some new filesystem objects.
         """
 
         cmd = self.repository.command("add", "--case-ok", "--not-recursive",
@@ -402,8 +437,6 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
         Commit the changeset.
         """
 
-        from sys import getdefaultencoding
-
         logmessage = []
 
         logmessage.append(date.strftime('%Y/%m/%d %H:%M:%S UTC'))
@@ -433,43 +466,24 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
 
         from os.path import join, exists
 
-        # darcs raises status 512 when it does not finding the entry,
+        # darcs raises status 512 when it does not find the entry,
         # removed by source. Since sometime a directory is left there
         # because it's not empty, darcs fails. So, do an explicit
         # remove on items that are still there.
 
         c = ExternalCommand(cwd=self.basedir,
                             command=self.repository.command("remove"))
-        c.execute([n for n in names if exists(join(self.basedir, n))])
-
+        existing = [n for n in names if exists(join(self.basedir, n))]
+        if existing:
+            c.execute(existing)
 
     def _renamePathname(self, oldname, newname):
         """
         Rename a filesystem object.
         """
 
-        from os.path import join, exists
-        from os import rename
-
-        # Check to see if the oldentry is still there. If it is,
-        # that probably means one thing: it's been moved and then
-        # replaced, see svn 'R' event. In this case, rename the
-        # existing old entry to something else to trick "darcs mv"
-        # (that will assume the move was already done manually) and
-        # finally restore its name.
-
-        absold = join(self.basedir, oldname)
-        renamed = exists(absold)
-        if renamed:
-            rename(absold, absold + '-TAILOR-HACKED-TEMP-NAME')
-
-        try:
-            cmd = self.repository.command("mv")
-            ExternalCommand(cwd=self.basedir, command=cmd).execute(oldname,
-                                                                   newname)
-        finally:
-            if renamed:
-                rename(absold + '-TAILOR-HACKED-TEMP-NAME', absold)
+        cmd = self.repository.command("mv")
+        ExternalCommand(cwd=self.basedir, command=cmd).execute(oldname, newname)
 
     def _prepareTargetRepository(self):
         """
@@ -500,13 +514,13 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
                             for md in IGNORED_METADIRS])
 
             # Eventually omit our own log...
-            logfile = self.repository.project.logfile
+            logfile = self.repository.projectref().logfile
             if logfile.startswith(self.basedir):
                 ignored.append('^%s$' %
                                escape(logfile[len(self.basedir)+1:]))
 
             # ... and state file
-            sfname = self.repository.project.state_file.filename
+            sfname = self.repository.projectref().state_file.filename
             if sfname.startswith(self.basedir):
                 sfrelname = sfname[len(self.basedir)+1:]
                 ignored.append('^%s$' % escape(sfrelname))
@@ -558,16 +572,15 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
                     skip = True
                     break
             if skip:
-                self.log_info('Entry %r skipped per boring rules' %
-                              e.name)
+                self.log.info('Entry "%s" skipped per boring rules', e.name)
                 skipped = True
             else:
                 entries.append(e)
 
         # All entries are gone, don't commit this changeset
         if not entries:
-            self.log_info('All entries ignored, skipping whole '
-                          'changeset %r' % changeset.revision)
+            self.log.info('All entries ignored, skipping whole '
+                          'changeset "%s"', changeset.revision)
             return None
 
         if skipped:
@@ -575,3 +588,47 @@ class DarcsWorkingDir(UpdatableSourceWorkingDir,SyncronizableTargetWorkingDir):
             adapted.entries = entries
 
         return adapted
+
+    def _tag(self, tag):
+        """
+        Apply the given tag to the repository, unless it has already
+        been applied to the current state. (If it has been applied to
+        an earlier state, do apply it; the later tag overrides the
+        earlier one.
+        """
+        if tag not in self._currentTags():
+            cmd = self.repository.command("tag", "--author", "Unknown tagger")
+            ExternalCommand(cwd=self.basedir, command=cmd).execute(tag)
+
+    def _currentTags(self):
+        """
+        Return a list of tags that refer to the repository's current
+        state.  Does not consider tags themselves to be part of the
+        state, so if the repo was tagged with T1 and then T2, then
+        both T1 and T2 are considered to refer to the current state,
+        even though 'darcs get --tag=T1' and 'darcs get --tag=T2'
+        would have different results (the latter creates a repo that
+        contains tag T2, but the former does not).
+
+        This function assumes that a tag depends on all patches that
+        precede it in the "darcs changes" list.  This assumption is
+        valid if tags only come into the repository via tailor; if the
+        user applies a tag by hand in the hybrid repository, or pulls
+        in a tag from another darcs repository, then the assumption
+        could be violated and mistagging could result.
+        """
+        cmd = self.repository.command("changes", "--from-match=not name ^TAG",
+                                      "--xml-output", "--reverse")
+        changes =  ExternalCommand(cwd=self.basedir, command=cmd)
+        output = changes.execute(stdout=PIPE, stderr=STDOUT)[0]
+        if changes.exit_status:
+            raise ChangesetApplicationFailure(
+                "%s returned status %d saying\n%s" %
+                (str(changes), changes.exit_status, output.read()))
+
+        tags = []
+        for cs in changesets_from_darcschanges_unsafe(output):
+            for tag in cs.tags:
+                if tag not in tags:
+                    tags.append(tag)
+        return tags
