@@ -17,7 +17,7 @@ from source import UpdatableSourceWorkingDir, InvocationError, \
      ChangesetApplicationFailure, GetUpstreamChangesetsFailure
 from target import SyncronizableTargetWorkingDir, TargetInitializationFailure
 from changes import ChangesetEntry,Changeset
-from os.path import exists, join, isdir
+from os.path import exists, join, isdir, split
 from os import renames, access, F_OK
 from string import whitespace
 
@@ -71,6 +71,7 @@ class MonotoneChangeset(Changeset):
 
         Changeset.__init__(self, revision=revision, date=None, author=None, log="")
         self.lin_ancestor = linearized_ancestor
+        self.real_ancestors = None
 
     def __eq__(self, other):
         return (self.revision == other.revision)
@@ -81,10 +82,12 @@ class MonotoneChangeset(Changeset):
     def __str__(self):
         s = [Changeset.__str__(self)]
         s.append('linearized ancestor: %s' % self.lin_ancestor)
-        s.append('real ancestor(s): %s' % ','.join(self.real_ancestors))
+        s.append('real ancestor(s): %s' %
+                 (self.real_ancestors and ','.join(self.real_ancestors)
+                  or 'None'))
         return '\n'.join(s)
 
-    def update(self, real_dates, authors, log, real_ancestors):
+    def update(self, real_dates, authors, log, real_ancestors, branches):
         """
         Updates the monotone changeset secondary data
         """
@@ -93,6 +96,7 @@ class MonotoneChangeset(Changeset):
         self.date = real_dates[0]
         self.real_dates = real_dates
         self.real_ancestors = real_ancestors
+        self.branches = branches
 
 class MonotoneLogParser:
     """
@@ -136,6 +140,7 @@ class MonotoneLogParser:
         self.authors=[]
         self.dates=[]
         self.changelog=""
+        self.branches=[]
 
         cmd = self.repository.command("log", "--db", self.repository.repository,
                                       "--last", "1", "--revision", revision)
@@ -173,7 +178,11 @@ class MonotoneLogParser:
                     date = datetime(y,m,d,hh,mm,ss)
                     self.dates.append(date)
                     state = self.SINGLE
-            elif pr("Branch:") or pr("Tag"):
+            elif pr("Branch:"):
+                # branch data
+                self.branches.append(pr.value)
+                state = self.SINGLE
+            elif pr("Tag"):
                 # unused data, just resetting state
                 state = self.SINGLE
             elif pr("Deleted files:") or pr("Deleted directories:"):
@@ -215,7 +224,8 @@ class MonotoneLogParser:
         chset.update(real_dates=self.dates,
                      authors=self.authors,
                      log=self.changelog,
-                     real_ancestors=self.ancestors)
+                     real_ancestors=self.ancestors,
+                     branches=self.branches)
 
         return chset
 
@@ -300,6 +310,12 @@ class MonotoneDiffParser:
         self.working_dir = working_dir
         self.repository = repository
 
+    def _addPathToSet(self, s, path):
+        parts = path.split('/')
+        while parts:
+            s.add('/'.join(parts))
+            parts.pop()
+
     def convertDiff(self, chset):
         """
         Fills a chset with the details data coming by a diff between
@@ -326,6 +342,9 @@ class MonotoneDiffParser:
             raise GetUpstreamChangesetsFailure(
                 "monotone diff returned status %d" % mtl.exit_status)
 
+        implicit_dirs_add = set()
+        dirs_add = set()
+
         # monotone diffs are prefixed by a section containing
         # metainformations about files
         # The section terminates with the first file diff, and each
@@ -350,9 +369,16 @@ class MonotoneDiffParser:
                             "followed by '%s'" %(token, fname))
 
                     # ok, is a file, control changesets data
-                    if token == "add_file" or token=="add_directory":
+                    if token == "add_file":
                         chentry = chset.addEntry(fname[1:-1], chset.revision)
                         chentry.action_kind = chentry.ADDED
+                        # split filespec in parts, skipping the filename
+                        # (monotone always uses '/' as path separator)
+                        self._addPathToSet(implicit_dirs_add, split(chentry.name)[0])
+                    elif token=="add_directory":
+                        chentry = chset.addEntry(fname[1:-1], chset.revision)
+                        chentry.action_kind = chentry.ADDED
+                        self._addPathToSet(dirs_add, chentry.name)
                     elif token == "delete_file" or token=="delete_directory":
                         chentry = chset.addEntry(fname[1:-1], chset.revision)
                         chentry.action_kind = chentry.DELETED
@@ -367,6 +393,8 @@ class MonotoneDiffParser:
                         chentry = chset.addEntry(newname[1:-1], chset.revision)
                         chentry.action_kind = chentry.RENAMED
                         chentry.old_name= fname[1:-1]
+                        if token=="rename_directory":
+                            self._addPathToSet(dirs_add, chentry.name)
                     elif token == "patch":
                         # patch entries are in the form: from oldrev to newrev
                         fromw = tkiter.next()
@@ -391,6 +419,13 @@ class MonotoneDiffParser:
             if in_item:
                 raise GetUpstreamChangesetsFailure("Unexpected end of 'diff' parsing changeset info")
 
+        # remove explicit dir adds from implicit set, then create "add" changesets
+        # for the remaining dirs
+        implicit_dirs_add = implicit_dirs_add.difference(dirs_add)
+        for d in implicit_dirs_add:
+            chentry = chset.addEntry(d, chset.revision)
+            chentry.action_kind = chentry.ADDED
+
 
 class MonotoneRevToCset:
     """
@@ -430,7 +465,7 @@ class MonotoneRevToCset:
       not used by tailor. Ignored
 
     branch
-      ignored (tailor follows only a single branch)
+      used to restrict source revs (tailor follows only a single branch)
 
     testresult
       ignored
@@ -451,20 +486,15 @@ class MonotoneRevToCset:
       linearized ancestor (i.e. previous revision in the linearized history)
     """
 
-    def __init__(self, repository, working_dir):
+    def __init__(self, repository, working_dir, branch):
         self.working_dir = working_dir
         self.repository = repository
+        self.branch = branch
         self.logparser = MonotoneLogParser(repository=repository,
                                            working_dir=working_dir)
         self.diffparser = MonotoneDiffParser(repository=repository,
                                              working_dir=working_dir)
 
-    def _cset_from_rev(self, lin_ancestor, revision):
-        # prepare a new changeset and fill it with rev data
-        chset = MonotoneChangeset(lin_ancestor, revision)
-        self.updateCset(chset)
-        return chset
-
     def updateCset(self, chset):
         # Parsing the log fills the changeset from revision data
         self.logparser.convertLog(chset)
@@ -473,20 +503,29 @@ class MonotoneRevToCset:
         if chset.lin_ancestor:
             self.diffparser.convertDiff(chset)
 
-    def getCset(self, revlist):
-        # receives a revlist, already toposorted (i.e. ordered by
-        # ancestry) and outputs a list of changesets
+    def getCset(self, revlist, onlyFirst):
+        """
+        receives a revlist, already toposorted (i.e. ordered by
+        ancestry) and outputs a list of changesets, filtering out revs
+        outside the chosen branch. If onlyFirst is true, only the
+        first valid element is considered
+        """
         cslist=[]
         anc=revlist[0]
         for r in revlist[1:]:
-            cslist.append(self._cset_from_rev(anc, r))
-            anc=r
+            chtmp = MonotoneChangeset(anc, r)
+            self.logparser.convertLog(chtmp)
+            if self.branch in chtmp.branches:
+                cslist.append(MonotoneChangeset(anc, r)) # using a new, unfilled changeset
+                anc=r
+                if onlyFirst:
+                    break
         return cslist
 
 
 class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDir):
 
-    def convert_head_initial(self, repository, module, revision, working_dir):
+    def _convert_head_initial(self, dbrepo, module, revision, working_dir):
         """
         This method handles HEAD and INITIAL pseudo-revisions, converting
         them to monotone revids
@@ -495,7 +534,7 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
         if revision == 'HEAD' or revision=='INITIAL':
             # in both cases we need the head(s) of the requested branch
             cmd = self.repository.command("automate","heads",
-                                          "--db", repository, module)
+                                          "--db", dbrepo, module)
             mtl = ExternalCommand(cwd=working_dir, command=cmd)
             outstr = mtl.execute(stdout=PIPE)
             if mtl.exit_status:
@@ -514,13 +553,13 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
                 # if the branch has multiple heads then we could end
                 # up with only part of the ancestry graph.
                 if len(revision)>1:
-                    self.log_info("Branch '%s' has multiple heads. There "
-                                  "is no guarantee to reconstruct the "
-                                  "full history." % module)
+                    self.log.info('Branch "%s" has multiple heads. There '
+                                  'is no guarantee to reconstruct the '
+                                  'full history.', module)
                 cmd = [ self.repository.command("automate","ancestors",
-                                                "--db",repository),
+                                                "--db",dbrepo),
                         self.repository.command("automate","toposort",
-                                                "--db",repository, "-@-")
+                                                "--db",dbrepo, "-@-")
                         ]
                 cmd[0].extend(revision)
                 cld = ExternalCommandChain(cwd=working_dir, command=cmd)
@@ -528,8 +567,18 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
                 if cld.exit_status:
                     raise InvocationError("Ancestor reading returned "
                                           "status %d" % cld.exit_status)
-                revision = outstr[0].getvalue().split()
-                effective_rev=revision[0]
+                revlist = outstr[0].getvalue().split()
+                if len(revlist)>1:
+                    mtr = MonotoneRevToCset(repository=self.repository,
+                                            working_dir=working_dir,
+                                            branch=module)
+                    first_cset = mtr.getCset(revlist, True)
+                    if len(first_cset)==0:
+                        raise InvocationError("Can't find an INITIAL revision on branch '%s'."
+                                              % module)
+                    effective_rev=first_cset[0].revision
+                else:
+                    effective_rev=revlist[0]
         return effective_rev
 
     ## UpdatableSourceWorkingDir
@@ -552,16 +601,16 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
                                   "status %d" % cld.exit_status)
 
         # now childs is a list of revids, we must transform it in a
-        # list of monotone changesets at this time we fill only the
+        # list of monotone changesets. We fill only the
         # linearized ancestor and revision ids, because at this time
         # we need only to know WICH changesets must be applied to the
-        # target repo, not WHAT are the changesets
-        childs = outstr[0].getvalue().split()
-        chlist = []
-        lin_anc=sincerev
-        for r in childs:
-            chlist.append(MonotoneChangeset(lin_anc, r))
-            lin_anc = r
+        # target repo, not WHAT are the changesets (apart for filtering
+        # the outside-branch revs)
+        childs = [sincerev] +outstr[0].getvalue().split()
+        mtr = MonotoneRevToCset(repository=self.repository,
+                                working_dir=self.repository.rootdir,
+                                branch=self.repository.module)
+        chlist = mtr.getCset(childs, False)
         return chlist
 
     def _applyChangeset(self, changeset):
@@ -572,7 +621,8 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
             raise ChangesetApplicationFailure("'mtn update' returned "
                                               "status %s" % mtl.exit_status)
         mtr = MonotoneRevToCset(repository=self.repository,
-                                working_dir=self.basedir)
+                                working_dir=self.basedir,
+                                branch=self.repository.module)
         mtr.updateCset( changeset )
 
         return False   # no conflicts
@@ -581,24 +631,27 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
         """
         Concretely do the checkout of the FIRST upstream revision.
         """
-        effrev = self.convert_head_initial(self.repository.repository,
+        effrev = self._convert_head_initial(self.repository.repository,
                                            self.repository.module, revision,
-                                           self.basedir)
+                                           self.repository.rootdir)
         if not exists(join(self.basedir, 'MT')):
-            self.log_info("checking out a working copy")
+
+            # actually check out the revision
+            self.log.info("Checking out a working copy")
             cmd = self.repository.command("co",
                                           "--db", self.repository.repository,
                                           "--revision", effrev,
                                           "--branch", self.repository.module,
-                                          self.repository.subdir)
+                                          self.basedir)
             mtl = ExternalCommand(cwd=self.repository.rootdir, command=cmd)
             mtl.execute()
             if mtl.exit_status:
                 raise TargetInitializationFailure(
                     "'monotone co' returned status %s" % mtl.exit_status)
         else:
-            self.log_info("%s already exists, assuming it's a monotone "
-                          "working dir" % self.basedir)
+            self.log.debug("%r already exists, assuming it's a monotone "
+                           "working dir already populated", self.basedir)
+
 
         # Ok, now the workdir contains the checked out revision. We
         # need to return a changeset describing it.  Since this is the
@@ -609,7 +662,8 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
         # now we update the new chset with basic data - without the
         # linearized ancestor, changeset entries will NOT be filled
         mtr = MonotoneRevToCset(repository=self.repository,
-                                working_dir=self.basedir)
+                                working_dir=self.basedir,
+                                branch=self.repository.module)
         mtr.updateCset(chset)
         return chset
 
@@ -623,8 +677,7 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
         fnames=[]
         for fn in names:
             if isdir(join(self.basedir, fn)):
-                self.log_info("ignoring addition of directory '%s' (%s)" %
-                              (fn, join(self.basedir, fn)) )
+                self.log.debug("ignoring addition of directory %r", fn)
             else:
                 fnames.append(fn)
         if len(fnames):
@@ -634,7 +687,7 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
             add.execute(fnames)
             if add.exit_status:
                 raise ChangesetApplicationFailure("%s returned status %s" %
-                                                  (str(add),add.exit_status))
+                                                    (str(add),add.exit_status))
 
 
     def _addSubtree(self, subdir):
@@ -653,9 +706,9 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
         Commit the changeset.
         """
 
-        from sys import getdefaultencoding
+        from locale import getpreferredencoding
 
-        encoding = ExternalCommand.FORCE_ENCODING or getdefaultencoding()
+        encoding = ExternalCommand.FORCE_ENCODING or getpreferredencoding()
 
         logmessage = []
         if patchname:
@@ -668,7 +721,8 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
         log.write('\n'.join(logmessage))
         log.close()
 
-        cmd = self.repository.command("commit", "--author", author,
+        cmd = self.repository.command("commit",
+                                      "--author", author.encode(encoding),
                                       "--date", date.isoformat(),
                                       "--message-file", rontf.name)
         commit = ExternalCommand(cwd=self.basedir, command=cmd)
@@ -682,12 +736,12 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
         # we ignore those errors ...
         if commit.exit_status:
             text = error.read()
-            if text.find("monotone: misuse: no changes to commit") == -1:
-                self.log_error(text)
+            if not "monotone: misuse: no changes to commit" in text:
+                self.log.error("Monotone commit said: %s", text)
                 raise ChangesetApplicationFailure(
                     "%s returned status %s" % (str(commit),commit.exit_status))
             else:
-                self.log_info("No changes to commit - changeset ignored")
+                self.log.info("No changes to commit - changeset ignored")
 
     def _removePathnames(self, names):
         """
@@ -696,18 +750,22 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
 
         # Monotone currently doesn't allow removing a directory, so we
         # must remove every item separately and intercept monotone
-        # directory errore messages.  We can't just filter the
+        # directory error messages.  We can't just filter the
         # directories, because the wc doesn't contain them anymore ...
         cmd = self.repository.command("drop")
         drop = ExternalCommand(cwd=self.basedir, command=cmd)
         for fn in names:
             dum, error = drop.execute(fn, stderr=PIPE)
             if drop.exit_status:
-                if not error.read().find("drop <directory>"):
-                    self.log_error(error.read())
+                errtext = error.read()
+                if not "drop <directory>" in errtext:
+                    self.log.error("Monotone drop said: %s", errtext)
                     raise ChangesetApplicationFailure("%s returned status %s" %
                                                       (str(drop),
                                                        drop.exit_status))
+                else:
+                    self.log.debug("Ignoring monotone whining about drop "
+                                   "of directory %r",  fn)
 
     def _renamePathname(self, oldname, newname):
         """
@@ -722,7 +780,7 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
                                                   "Both names already exist" %
                                                   (oldname, newname))
             renames(join(self.basedir, newname), join(self.basedir, oldname))
-            self.log_info("preparing to rename %s->%s" % (oldname, newname))
+            self.log.debug('Renamed "%s" back to "%s"', newname, oldname)
 
         cmd = self.repository.command("rename")
         rename = ExternalCommand(cwd=self.basedir, command=cmd)
@@ -749,43 +807,43 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
                                               "the monotone db at %r" %
                                               target_repository)
 
-        if target_repository.keyfile:
-            # a key file is available, read into the database
-            keyfile = file(target_repository.keyfile)
-            cmd = self.repository.command("read", "--db",
-                                          target_repository.repository)
-            regkey = ExternalCommand(command=cmd)
-            regkey.execute(input=keyfile)
+        if target_repository.keyid:
+            self.repository.log_info("Using key %s for commits" % (target_repository.keyid,))
         else:
-            # no keyfile specified, generate a new key - if a
-            # passphrase is defined, automatically provide it The
-            # keyid must be available
-            if not target_repository.keyid:
+            # keystore key id unspecified, look at other options
+            if target_repository.keyfile:
+                # a key file is available, read into the database
+                keyfile = file(target_repository.keyfile)
+                cmd = self.repository.command("read", "--db",
+                                              target_repository.repository)
+                regkey = ExternalCommand(command=cmd)
+                regkey.execute(input=keyfile)
+            elif target_repository.keygenid:
+                # requested a new key
+                cmd = self.repository.command("genkey", "--db",
+                                              target_repository.repository)
+                regkey = ExternalCommand(command=cmd)
+                if target_repository.passphrase:
+                    passp="%s\n%s\n" % (target_repository.passphrase,
+                                        target_repository.passphrase)
+                regkey.execute(target_repository.keygenid, input=passp)
+            else:
                 raise TargetInitializationFailure("Can't setup the monotone "
                                                   "repository %r. "
-                                                  "A keyfile or keyid must "
-                                                  "be provided." %
+                                                  "A keyid, keyfile or keygenid "
+                                                  "must be provided." %
                                                   target_repository)
-            cmd = self.repository.command("genkey", "--db",
-                                          target_repository.repository)
-            regkey = ExternalCommand(command=cmd)
-            if target_repository.passphrase:
-                passp="%s\n%s\n" % (target_repository.passphrase,
-                                    target_repository.passphrase)
-            regkey.execute(target_repository.keyid, input=passp)
 
-        if regkey.exit_status:
-            raise TargetInitializationFailure("Was not able to setup "
-                                              "the monotone initial key at %r" %
-                                              target_repository)
+            if regkey.exit_status:
+                raise TargetInitializationFailure("Was not able to setup "
+                                                  "the monotone initial key at %r" %
+                                                  target_repository)
 
     def _prepareTargetRepository(self):
         """
         Check for target repository existence, eventually create it.
         """
 
-        from os.path import exists
-
         if not self.repository.repository:
             return
 
@@ -798,20 +856,22 @@ class MonotoneWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDi
         upstream source tree, when overriden by subclasses.
         """
 
-        from os.path import join, exists
-
         if not self.repository.repository or exists(join(self.basedir, 'MT')):
             return
 
-        cmd = self.repository.command("setup",
-                                      "--db", self.repository.repository,
-                                      "--branch", self.repository.module)
-
         if not self.repository.module:
             raise TargetInitializationFailure("Monotone needs a module "
                                               "defined (to be used as "
                                               "commit branch)")
 
+
+        cmd = self.repository.command("setup",
+                                      "--db", self.repository.repository,
+                                      "--branch", self.repository.module)
+
+        if self.repository.keyid:
+            cmd.extend( ("--key", self.repository.keyid) )
+
         setup = ExternalCommand(command=cmd)
         setup.execute(self.basedir)
 
