@@ -15,13 +15,13 @@ from shwrap import ExternalCommand, PIPE, STDOUT, ReopenableNamedTemporaryFile
 from source import UpdatableSourceWorkingDir, \
      ChangesetApplicationFailure, GetUpstreamChangesetsFailure
 from target import SyncronizableTargetWorkingDir, TargetInitializationFailure
+from config import ConfigurationError
 
-def changesets_from_svnlog(log, repository, module):
-    from xml.sax import parseString
-    from xml.sax.handler import ContentHandler
+def changesets_from_svnlog(log, repository, module, chunksize=2**15):
+    from xml.sax import make_parser
+    from xml.sax.handler import ContentHandler, ErrorHandler
     from changes import ChangesetEntry, Changeset
     from datetime import datetime
-    from string import maketrans
 
     def get_entry_from_path(path, module=module):
         # Given the repository url of this wc, say
@@ -68,6 +68,7 @@ def changesets_from_svnlog(log, repository, module):
             self.current = None
             self.current_field = []
             self.renamed = {}
+            self.external_copies = []
 
         def startElement(self, name, attributes):
             if name == 'logentry':
@@ -99,17 +100,34 @@ def changesets_from_svnlog(log, repository, module):
                 # at the same time change related entry from ADDED to
                 # RENAMED.
 
+                # When copying a directory from another location in the
+                # repository (outside the tracked tree), SVN will report files
+                # below this dir that are not being committed as being
+                # removed.
+
+                # We thus need to change the action_kind for all entries
+                # that are below a dir that was "copyfrom" from a path
+                # outside of this module:
+                #  D -> Remove entry completely (it's not going to be in here)
+                #  (M,A,R) -> A
+
                 mv_or_cp = {}
                 for e in self.current['entries']:
                     if e.action_kind == e.ADDED and e.old_name is not None:
                         mv_or_cp[e.old_name] = e
 
+                def parent_was_copied_externally(n):
+                    for p in self.external_copies:
+                        if n.startswith(p):
+                            return True
+                    return False
+
                 entries = []
                 for e in self.current['entries']:
                     if e.action_kind==e.DELETED and mv_or_cp.has_key(e.name):
                         mv_or_cp[e.name].action_kind = e.RENAMED
                     elif e.action_kind=='R':
-                        # In svn parlance, 'R' means Replaced: a tipical
+                        # In svn parlance, 'R' means Replaced: a typical
                         # scenario is
                         #   $ svn mv a.txt b.txt
                         #   $ touch a.txt
@@ -118,6 +136,10 @@ def changesets_from_svnlog(log, repository, module):
                             mv_or_cp[e.name].action_kind = e.RENAMED
                         e.action_kind = e.ADDED
                         entries.append(e)
+                    elif parent_was_copied_externally(e.name):
+                        if e.action_kind != e.DELETED:
+                            e.action_kind = e.ADDED
+                            entries.append(e)
                     else:
                         entries.append(e)
 
@@ -150,27 +172,31 @@ def changesets_from_svnlog(log, repository, module):
                             entry.old_name = old
                             self.renamed[entry.old_name] = True
                         else:
+                            self.external_copies.append(entry.name)
                             entry.action_kind = entry.ADDED
                     else:
                         entry.action_kind = self.ACTIONSMAP[self.current_path_action]
 
                     self.current['entries'].append(entry)
 
-
         def characters(self, data):
             self.current_field.append(data)
 
-
-    # Apparently some (SVN repo contains)/(SVN server dumps) some characters that
-    # are illegal in an XML stream. This was the case with Twisted Matrix master
-    # repository. To be safe, we replace all of them with a question mark.
-
-    allbadchars = "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0B\x0C\x0E\x0F\x10\x11" \
-                  "\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F\x7f"
-    tt = maketrans(allbadchars, "?"*len(allbadchars))
+    parser = make_parser()
     handler = SvnXMLLogHandler()
-    parseString(log.read().translate(tt), handler)
-    return handler.changesets
+    parser.setContentHandler(handler)
+    parser.setErrorHandler(ErrorHandler())
+
+    chunk = log.read(chunksize)
+    while chunk:
+        parser.feed(chunk)
+        for cs in handler.changesets:
+            yield cs
+        handler.changesets = []
+        chunk = log.read(chunksize)
+    parser.close()
+    for cs in handler.changesets:
+        yield cs
 
 
 class SvnWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDir):
@@ -191,6 +217,25 @@ class SvnWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDir):
         if svnlog.exit_status:
             return []
 
+        if self.repository.filter_badchars:
+            from string import maketrans
+            from cStringIO import StringIO
+
+            # Apparently some (SVN repo contains)/(SVN server dumps) some
+            # characters that are illegal in an XML stream. This was the case
+            # with Twisted Matrix master repository. To be safe, we replace
+            # all of them with a question mark.
+
+            if isinstance(self.repository.filter_badchars, string):
+                allbadchars = self.repository.filter_badchars
+            else:
+                allbadchars = "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09" \
+                              "\x0B\x0C\x0E\x0F\x10\x11\x12\x13\x14\x15" \
+                              "\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F\x7f"
+
+            tt = maketrans(allbadchars, "?"*len(allbadchars))
+            log = StringIO(log.read().translate(tt))
+
         return changesets_from_svnlog(log,
                                       self.repository.repository,
                                       self.repository.module)
@@ -204,32 +249,34 @@ class SvnWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDir):
 
         retry = 0
         while True:
-            out = svnup.execute(stdout=PIPE)[0]
+            out, err = svnup.execute(stdout=PIPE, stderr=PIPE)
 
             if svnup.exit_status == 1:
                 retry += 1
                 if retry>3:
                     break
                 delay = 2**retry
-                self.log_info("%s returned status %s, "
-                              "retrying in %d seconds..." %
-                              (str(svnup), svnup.exit_status, delay))
+                self.log.error("%s returned status %s saying\n%s",
+                               str(svnup), svnup.exit_status, err.read())
+                self.log.warning("Retrying in %d seconds...", delay)
                 sleep(delay)
             else:
                 break
 
         if svnup.exit_status:
             raise ChangesetApplicationFailure(
-                "%s returned status %s" % (str(svnup), svnup.exit_status))
+                "%s returned status %s saying\n%s" % (str(svnup),
+                                                     svnup.exit_status,
+                                                     err.read()))
 
-        self.log_info("%s updated to %s" % (
-            ','.join([e.name for e in changeset.entries]),
-            changeset.revision))
+        self.log.debug("%s updated to %s",
+                       ','.join([e.name for e in changeset.entries]),
+                       changeset.revision)
 
         result = []
         for line in out:
             if len(line)>2 and line[0] == 'C' and line[1] == ' ':
-                self.log_info("Conflict after 'svn update': '%s'" % line)
+                self.log.warning("Conflict after svn update: %r", line)
                 result.append(line[2:-1])
 
         return result
@@ -241,59 +288,94 @@ class SvnWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDir):
 
         from os.path import join, exists
 
+        # Verify that the we have the root of the repository: do that
+        # iterating an "svn ls" over the hierarchy until one fails
+
+        lastok = self.repository.repository
+        if not self.repository.trust_root:
+            cmd = self.repository.command("ls")
+            svnls = ExternalCommand(command=cmd)
+            svnls.execute(self.repository.repository)
+
+            reporoot = lastok[:lastok.rfind('/')]
+            # Don't go too far, that is, stop when you hit schema://...
+            while '//' in reporoot:
+                svnls.execute(reporoot)
+                if svnls.exit_status:
+                    break
+                lastok = reporoot
+                reporoot = reporoot[:reporoot.rfind('/')]
+
+        if lastok <> self.repository.repository:
+            module = self.repository.repository[len(lastok):]
+            module += self.repository.module
+            raise ConfigurationError("Non-root svn repository %r. "
+                                     "Please specify that as 'repository=%s' "
+                                     "and 'module=%s'." %
+                                     (self.repository.repository,
+                                      lastok, module.rstrip('/')))
+
         if revision == 'INITIAL':
             initial = True
             cmd = self.repository.command("log", "--verbose", "--xml",
-                                          "--limit", "1",
+                                          "--stop-on-copy",
                                           "--revision", "1:HEAD")
+            if self.repository.use_limit:
+                cmd.extend(["--limit", "1"])
             svnlog = ExternalCommand(command=cmd)
-            output = svnlog.execute("%s%s" % (self.repository.repository,
-                                              self.repository.module),
-                                    stdout=PIPE)[0]
+            out, err = svnlog.execute("%s%s" % (self.repository.repository,
+                                                self.repository.module),
+                                      stdout=PIPE, stderr=PIPE)
 
             if svnlog.exit_status:
                 raise TargetInitializationFailure(
-                    "%s returned status %d saying \"%s\"" %
-                    (str(output), changes.exit_status, output.read()))
+                    "%s returned status %d saying\n%s" %
+                    (str(svnlog), svnlog.exit_status, err.read()))
 
-            csets = changesets_from_svnlog(output,
+            csets = changesets_from_svnlog(out,
                                            self.repository.repository,
                                            self.repository.module)
-            revision = csets[0].revision
+            last = csets.next()
+            revision = last.revision
         else:
             initial = False
 
         if not exists(join(self.basedir, '.svn')):
-            self.log_info("checking out a working copy")
+            self.log.debug("Checking out a working copy")
             cmd = self.repository.command("co", "--quiet",
                                           "--revision", revision)
             svnco = ExternalCommand(command=cmd)
-            svnco.execute("%s%s" % (self.repository.repository,
-                                    self.repository.module), self.basedir)
+            out, err = svnco.execute("%s%s" % (self.repository.repository,
+                                               self.repository.module),
+                                     self.basedir, stdout=PIPE, stderr=PIPE)
             if svnco.exit_status:
                 raise TargetInitializationFailure(
-                    "%s returned status %s" % (str(svnco), svnco.exit_status))
+                    "%s returned status %s saying\n%s" % (str(svnco),
+                                                         svnco.exit_status,
+                                                         err.read()))
         else:
-            self.log_info("%s already exists, assuming it's a svn working dir" % self.basedir)
+            self.log.debug("%r already exists, assuming it's "
+                           "a svn working dir", self.basedir)
 
         if not initial:
+            if revision=='HEAD':
+                revision = 'COMMITTED'
             cmd = self.repository.command("log", "--verbose", "--xml",
-                                          "--revision", revision=='HEAD' and 'COMMITTED' or revision)
+                                          "--revision", revision)
             svnlog = ExternalCommand(cwd=self.basedir, command=cmd)
-            output = svnlog.execute(stdout=PIPE)[0]
+            out, err = svnlog.execute(stdout=PIPE, stderr=PIPE)
 
             if svnlog.exit_status:
                 raise TargetInitializationFailure(
-                    "%s returned status %d saying \"%s\"" %
-                    (str(changes), changes.exit_status, output.read()))
+                    "%s returned status %d saying\n%s" %
+                    (str(changes), changes.exit_status, err.read()))
 
-            csets = changesets_from_svnlog(output,
+            csets = changesets_from_svnlog(out,
                                            self.repository.repository,
                                            self.repository.module)
+            last = csets.next()
 
-        last = csets[0]
-
-        self.log_info("working copy up to svn revision %s" % last.revision)
+        self.log.debug("Working copy up to svn revision %s", last.revision)
 
         return last
 
@@ -326,9 +408,9 @@ class SvnWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDir):
         Commit the changeset.
         """
 
-        from sys import getdefaultencoding
+        from locale import getpreferredencoding
 
-        encoding = ExternalCommand.FORCE_ENCODING or getdefaultencoding()
+        encoding = ExternalCommand.FORCE_ENCODING or getpreferredencoding()
 
         logmessage = []
         if patchname:
@@ -355,11 +437,13 @@ class SvnWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDir):
         if not entries:
             entries = ['.']
 
-        out = commit.execute(entries, stdout=PIPE, LANG='C')[0]
+        out, err = commit.execute(entries, stdout=PIPE, stderr=PIPE, LANG='C')
 
         if commit.exit_status:
-            raise ChangesetApplicationFailure("%s returned status %d" %
-                                              (str(commit), commit.exit_status))
+            raise ChangesetApplicationFailure("%s returned status %d saying\n%s"
+                                              % (str(commit),
+                                                 commit.exit_status,
+                                                 err.read()))
         line = out.readline()
         if not line:
             # svn did not find anything to commit
@@ -387,7 +471,7 @@ class SvnWorkingDir(UpdatableSourceWorkingDir, SyncronizableTargetWorkingDir):
             propset = ExternalCommand(cwd=self.basedir, command=cmd)
 
             propset.execute(date.isoformat()+".000000Z", propname='svn:date')
-            propset.execute(author, propname='svn:author')
+            propset.execute(author.encode(encoding), propname='svn:author')
 
         cmd = self.repository.command("update", "--quiet",
                                       "--revision", revision)
